#include "config.h"

#include "global_options.hpp"

#include "config.hpp"

#include "gettext.hpp"

#include "boost.hpp"

#include <string>
#include <vector>

#include <iostream>

#include <cassert>

using std::string;
using std::vector;
using std::list;

typedef std::pair<string, string> string_pair;

pthread_mutex_t GlobalOptions::singleton_mutex = PTHREAD_MUTEX_INITIALIZER;

GlobalOptions::GlobalOptions()
{
  Config *conf = S_Config::get_instance();

  vector<string> cd_device_values;

  assert(conf->p_cdpath().size() == conf->p_cdname().size());

  list<string_pair > cd_names = conf->p_cdname();
  list<string_pair > cd_paths = conf->p_cdpath();

  // sanity check
  foreach (string_pair& path_pair, cd_paths) {
    bool found_match = false;
    foreach (string_pair& name_pair, cd_names)
      if (name_pair.first == path_pair.first) {
	found_match = true;
	break;
      }
    if (!found_match) {
      print_critical(gettext("name mismatch in cd configuration, please check your config"));
      exit(0);
    }
  }

  foreach (string_pair& name_pair, cd_names)
    cd_device_values.push_back(name_pair.second);

  cd_device = new Option(false, gettext("cd device"), "cd device",
			 0, cd_device_values, cd_device_values);

  vector<string> audio_device_values;
  audio_device_values.push_back(conf->p_first_dev_name());

  if (!conf->p_second_dev_name().empty())
    audio_device_values.push_back(conf->p_second_dev_name());

  audio_device = new Option(false, gettext("audio device"), "audio device",
			    0, audio_device_values, audio_device_values);
}

Option *GlobalOptions::dir_order()
{
  vector<string> dir_order_values;
  dir_order_values.push_back(gettext("directories first"));
  dir_order_values.push_back(gettext("by name"));

  vector<string> english_dir_order_values;
  english_dir_order_values.push_back("directories first");
  english_dir_order_values.push_back("by name");

  return new Option(false, gettext("sorting"), "sorting",
		    0, dir_order_values, english_dir_order_values);
}

GlobalOptions::~GlobalOptions()
{
  delete cd_device;
  delete audio_device;
}
