#include "libfspp.hpp"

#include "movie-library.hpp"

#include "renderer.hpp"
#include "movie_config.hpp"
#include "config.hpp"

#include "updater.hpp"

#include "mplayer.hpp"

#include "gsimplefile.hpp"

#include "cpp_sqlitepp.hpp"

#include "plugins.hpp"

#include "boost.hpp"

// --version
#include <getopt.h>

#ifdef use_nls
#include <langinfo.h>
#include <locale.h>
#endif

#include <vector>
#include <stack>
#include <list>
#include <string>

using std::vector;
using std::list;
using std::string;

void insert_file_into_db(const string& filename, const string& parent, SQLDatabase& db)
{
  std::cout << "checking filename:" << filename << std::endl;

  bool is_dir = isDirectory(filename);

  string good_filename;
  string cut_filename;

  if (is_dir) {
    good_filename = string_format::unique_folder_name(filename);
    cut_filename = good_filename.substr(0, good_filename.size()-1);
  } else {
    good_filename = filename;
    cut_filename = filename;
  }

  // check if already exists
  SQLQuery *q = db.query("Folders", ("SELECT id FROM %t WHERE filename='" + string_format::escape_db_string(good_filename) + "'").c_str());
  if (q && q->numberOfTuples() > 0) {
    delete q;
    return;
  } else
    delete q;

  string parent_id = "";

  string good_parent = string_format::unique_folder_name(parent);

  q = db.query("Folders", ("SELECT id FROM %t WHERE filename='" + string_format::escape_db_string(good_parent) + "'").c_str());
  if (q && q->numberOfTuples() > 0) {
    assert(q->numberOfTuples() == 1);
    SQLRow &row = (*q)[0];
    parent_id = row["id"];
  }
  delete q;

  if (parent_id.empty()) {
    char *tmp = sqlite3_mprintf("INSERT INTO Folders VALUES(NULL, '0', '%q', '', '1')", good_parent.c_str());
    db.execute(tmp);
    sqlite3_free(tmp);

    q = db.query("Folders", ("SELECT id FROM %t WHERE filename='" + string_format::escape_db_string(good_parent) + "'").c_str());
    if (q && q->numberOfTuples() > 0) {
      assert(q->numberOfTuples() == 1);
      SQLRow &row = (*q)[0];
      parent_id = row["id"];
    }
    delete q;
  }

  assert(!parent_id.empty());

  string::size_type i;
  if ((i = cut_filename.rfind('/')) == string::npos)
    i = 0;
  string name = cut_filename.substr((i != 0) ? i+1 : i);

  if (!is_dir && (i = name.rfind('.')) != string::npos)
    name = name.substr(0, i);

  char *tmp_insert = sqlite3_mprintf("INSERT INTO Folders VALUES(NULL, '%q', '%q', '%q', '%q')",
				     parent_id.c_str(), good_filename.c_str(), name.c_str(),
				     conv::itos(is_dir).c_str());
  db.execute(tmp_insert);
  sqlite3_free(tmp_insert);
}

bool create_thumbnail(const string& file, const string& thumbnail, int width, int height, Mplayer& thumbnailer)
{
  return thumbnailer.create_thumbnail(file, thumbnail, width, height);
}

void rdir_p(const string& argv, SQLDatabase& db)
{
  Config *conf = S_Config::get_instance();
  MovieConfig *movie_conf = S_MovieConfig::get_instance();
  Render *render = S_Render::get_instance();

  //  Plugins *plugins = S_Plugins::get_instance();
  //  MoviePlayerPlugin *thumbnailer = (MoviePlayerPlugin*)plugins->load(string(MMS_PLUGINS_DIR) + "/lib_feature_movie_player_mplayer.so");

  Mplayer thumbnailer;

  int info_width = 150, info_height = 150;

  int image_width = 125, image_height = 96;

  if (conf->p_v_res() == 405) {
    image_width = 80;
    image_height = 67;
  }

  std::stack<string> queued_dirs;

  std::cout << "starting to generate thumbnails for movie files" << std::endl;

  queued_dirs.push(argv);
  while (!queued_dirs.empty()) {
    string cur_dir = queued_dirs.top();
    queued_dirs.pop();
    for (file_iterator<file_t, default_order> i (cur_dir); i != i.end (); i.advance(false)) {

      string cur_file = string_format::remove_doubles(i->getName());

      if (isDirectory(cur_file)) {
	cur_file  = string_format::unique_folder_name(cur_file);
	insert_file_into_db(cur_file, cur_dir, db);
	queued_dirs.push(cur_file);
      } else {
	if (check_type(cur_file, movie_conf->p_filetypes_m()) == emptyMyPair)
	  continue;

	insert_file_into_db(cur_file, cur_dir, db);

	string thumbnail = "";

	string::size_type dot_pos = cur_file.rfind(".");
	if (dot_pos != string::npos) {
	  // list view
	  thumbnail = conf->p_temp_image_loc() + cur_file.substr(0, dot_pos) + "_" + conv::itos(info_width) + "_" + conv::itos(info_height) + "_thumb.jpg";
	  // try thumbnail
	  if (!file_exists(thumbnail))
	    create_thumbnail(cur_file, thumbnail, info_width, info_height, thumbnailer);

	  // normal
	  thumbnail = conf->p_temp_image_loc() + cur_file.substr(0, dot_pos) + "_" + conv::itos(image_width) + "_" + conv::itos(image_height) + "_thumb.jpg";
	  // try thumbnail
	  if (!file_exists(thumbnail))
	    create_thumbnail(cur_file, thumbnail, image_width, image_height, thumbnailer);
	}
      }
    }
  }
}

void create_indexes(SQLDatabase& db)
{
  db.execute("CREATE INDEX idx_folders_name ON Folders (name)");
  db.execute("CREATE INDEX idx_folders_parent ON Folders (parent)");
  db.execute("CREATE INDEX idx_folders_filename ON Folders (filename)");

  db.execute("CREATE INDEX idx_hdmovie_path ON HDMovie (path)");
}

void create_db(SQLDatabase& db)
{
  vector<string> fields;
  fields.push_back("version INTEGER NOT NULL");
  db.createTable("Schema", fields);

  db.execute("INSERT INTO Schema VALUES('2')");

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("parent INTEGER");
  fields.push_back("filename TEXT");
  fields.push_back("name TEXT");
  fields.push_back("is_folder INTEGER"); // 1 = true
  db.createTable("Folders", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("name TEXT");
  db.createTable("Directors", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("did INTEGER NOT NULL");
  fields.push_back("mid INTEGER NOT NULL");
  db.createTable("DMovie", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("name TEXT");
  fields.push_back("role TEXT");
  db.createTable("WritingCredits", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("wcid INTEGER NOT NULL");
  fields.push_back("mid INTEGER NOT NULL");
  db.createTable("WCMovie", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("name TEXT");
  db.createTable("Genre", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("gid INTEGER NOT NULL");
  fields.push_back("mid INTEGER NOT NULL");
  db.createTable("GMovie", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("name TEXT");
  fields.push_back("role TEXT");
  db.createTable("Actors", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("aid INTEGER NOT NULL");
  fields.push_back("mid INTEGER NOT NULL");
  db.createTable("AMovie", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("title TEXT");
  fields.push_back("path TEXT");
  fields.push_back("url TEXT");
  fields.push_back("runtime TEXT");
  fields.push_back("tagline TEXT");
  fields.push_back("plot TEXT");
  fields.push_back("rating REAL");
  fields.push_back("votes INTEGER");
  fields.push_back("year INTEGER");
  fields.push_back("top250 INTEGER");

  db.createTable("HDMovie", fields);

  create_indexes(db);
}

void check_db(SQLDatabase& db)
{
  if (db.hasTable("Schema")) {
    // check for correct version
    SQLQuery *q = db.query("Schema", "SELECT version FROM %t");
    if (q && q->numberOfTuples() > 0) {
      SQLRow &row = (*q)[0];
      if (row["version"] == "1") {
	db.execute("UPDATE Schema SET version='2' WHERE version='1'");

	create_indexes(db);
      }
    }
    delete q;
  } else {
    // create the whole db
    create_db(db);
  }
}

// stupid workaround
void Mms::clean_up()
{}

static struct option longopts[] =
{
  { "conf", 1, 0, 'c' },
  { "personal", 1, 0, 'p' },
  { "help", 0, 0, 'h' },
};

void set_default(Config *conf)
{
  string homedir = "";

  havemydir("", &homedir);
  conf->s_var_data_dir(homedir);
}

void set_config(Config *conf, string config)
{
  if (config[config.size()-1] != '/')
    config += '/';

  conf->s_homedir(config);
}

void set_personal(Config *conf, string personal)
{
  if (personal[personal.size()-1] != '/')
    personal += '/';

  conf->s_var_data_dir(personal);
}

void check_path(const string& argv, SQLDatabase& db)
{
  Config *conf = S_Config::get_instance();
  MovieConfig *movie_conf = S_MovieConfig::get_instance();
  Render *render = S_Render::get_instance();

  //  Plugins *plugins = S_Plugins::get_instance();
  //  MoviePlayerPlugin *thumbnailer = (MoviePlayerPlugin*)plugins->load(string(MMS_PLUGINS_DIR) + "/lib_feature_movie_player_mplayer.so");

  Mplayer thumbnailer;

  int info_width = 150, info_height = 150;

  int image_width = 125, image_height = 96;

  if (conf->p_v_res() == 405) {
    image_width = 80;
    image_height = 67;
  }

  std::stack<string> queued_dirs;

  std::cout << "starting to generate thumbnails for movie files" << std::endl;

  list<string> dirs = movie_conf->p_movie_dirs(); // make compiler happy
  foreach (string& dir, dirs)
    queued_dirs.push(dir);

  while (!queued_dirs.empty()) {
    string cur_dir = queued_dirs.top();
    queued_dirs.pop();

    if (argv.find(cur_dir) == string::npos)
      continue;

    for (file_iterator<file_t, default_order> i (cur_dir); i != i.end (); i.advance(false)) {

      string cur_file = string_format::remove_doubles(i->getName());

      if (isDirectory(cur_file)) {
	cur_file  = string_format::unique_folder_name(cur_file);
	insert_file_into_db(cur_file, cur_dir, db);
	queued_dirs.push(cur_file);
      } else {
	if (check_type(cur_file, movie_conf->p_filetypes_m()) == emptyMyPair)
	  continue;

	insert_file_into_db(cur_file, cur_dir, db);

	string thumbnail = "";

	string::size_type dot_pos = cur_file.rfind(".");
	if (dot_pos != string::npos) {
	  // list view
	  thumbnail = conf->p_temp_image_loc() + cur_file.substr(0, dot_pos) + "_" + conv::itos(info_width) + "_" + conv::itos(info_height) + "_thumb.jpg";
	  // try thumbnail
	  if (!file_exists(thumbnail))
	    create_thumbnail(cur_file, thumbnail, info_width, info_height, thumbnailer);

	  // normal
	  thumbnail = conf->p_temp_image_loc() + cur_file.substr(0, dot_pos) + "_" + conv::itos(image_width) + "_" + conv::itos(image_height) + "_thumb.jpg";
	  // try thumbnail
	  if (!file_exists(thumbnail))
	    create_thumbnail(cur_file, thumbnail, image_width, image_height, thumbnailer);
	}
      }
    }
  }
}

int main(int argc,char* argv[])
{
  Config *config = S_Config::get_instance();

  string homedir = "", var_data = "";

#ifdef use_nls
  // gettext
  setlocale(LC_ALL, ""); /* set from the environment variables */
  bindtextdomain("mms", mms_prefix"/share/locale");
  bind_textdomain_codeset("mms", nl_langinfo(CODESET));
  textdomain("mms");
#endif

  set_config(config, mms_config); // config
  set_default(config); // personal

  int c;
  int min_argc = 1;

  while ((c = getopt_long(argc, argv, "abc:dfhlnop:w:r:i:t:O:vV",
			  longopts, NULL)) != -1)
    switch (c) {
    case 'c': // config
      min_argc += 2;
      homedir = optarg;
      set_config(config, homedir);
      break;
    case 'p': // personal
      min_argc += 2;
      var_data = optarg;
      set_personal(config, var_data);
      break;
    case 'h':
      Print print (Print::INFO);
      print.add_line(gettext("My Media System movie library creation tool"));
      print.add_line("");
      print.add_line(gettext("Usage: ./mms-movie-library <directory> <directory> ..."));
      print.add_line("");
      print.add_line(gettext("Options:"));
      print.add_line(gettext(" -c <absolute> : use alternative config directory"));
      print.add_line(gettext(" -p <absolute> : use alternative personal directory (db, playlist)"));
      print.print();
      exit(0);
      break;
    }

  Print print(gettext("Using ") + config->p_homedir() + gettext(" as configuration directory"), Print::INFO);
  Print print2(gettext("Using ") + config->p_var_data_dir() + gettext(" as personal directory"), Print::INFO);

  // now pass the configuration file using the correct path
  config->parse_configuration_file();

  S_MovieConfig::get_instance()->parse_configuration_file(config->p_homedir());

  if (argc <= min_argc) {
    std::cerr << "please specify at least one directory" << std::endl;
    return 1;
  }

  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();
  screen_updater->start();
  screen_updater->activate();

  BackgroundUpdater *background_updater = S_BackgroundUpdater::get_instance();

  background_updater->start();
  background_updater->activate();

  SQLDatabase db((config->p_var_data_dir() + "moviehd.db").c_str());

  check_db(db);

  for (int i = 1; i < argc; ++i) {
    // check that all the folders from root in config to dir list parsed is indexed
    check_path(argv[i], db);
    rdir_p(argv[i], db);
  }
}

