#include "config.h"

#include "options.hpp"

#include "common.hpp"
#include "graphics.hpp"
#include "global.hpp"

#include "plugins/plugins.hpp"

#include "gettext.hpp"
#include "shutdown.hpp"
#include "touch.hpp"

#include "boost.hpp"

#include <fstream>

// mkdir()
#include <sys/stat.h>
#include <sys/types.h>

using std::list;
using std::vector;
using std::string;

Options::Options()
  : exit_loop(false)
{
  im = S_InputMaster::get_instance();
  themes = S_Themes::get_instance();
  render = S_Render::get_instance();
  conf = S_Config::get_instance();
  go = S_GlobalOptions::get_instance();
  global = S_Global::get_instance();
}

bool Options::mainloop()
{
  im->save_map();
  im->set_map("default");

  Shutdown *sd = S_Shutdown::get_instance();

  Input input;

  vector<Option*>::iterator iter = val.begin();
  int pos = 0; // position in values, used by print()

  // set the element to the first non-type element
  while ((*iter)->type) {
    vector<Option*>::iterator j = val.end(); --j;
    if (iter != j) {
      ++iter;
      ++pos;
    } else {
      iter = val.begin();
      pos = 0;
    }
  }

  bool reload = false, reconfigure = false;

  bool update_needed = true;

  Render *render = S_Render::get_instance();

  int curlayer = render->device->get_current_layer();
  render->device->switch_to_layer(curlayer+1);
  render->device->animation_section_begin();
  render->device->layer_active_no_wait(true);
  render->device->reset_layout_attribs_nowait();
  render->device->animation_move( 0, -conf->p_v_res(), 0, 0, 60, curlayer+1);
  render->device->animation_move(0, 0, 0, conf->p_v_res(), 60, curlayer);
  render->device->animation_fade(0,1,100,curlayer+1);
  render->device->animation_fade(1,0,100,curlayer);
  render->device->animation_section_end();

  while (!exit_loop)
    {
      if (update_needed) {
	print((*iter)->name, pos);
	print_lcd_menu((*iter)->name);
      }

      input = im->device->get_input();

      if (sd->is_enabled()) {
      	sd->cancel();
      	continue;
      }

      update_needed = true;

      if (input.key == "touch_input") {
	S_Touch::get_instance()->run_callback();
	continue;
      }

      if (input.command == "prev")
 	{
	  do {
	    if (iter != val.begin()) {
	      --iter; --pos;
	    } else {
	      vector<Option*>::iterator j = val.end(); --j;
	      iter = j; pos = val.size() - 1;
	    }
	  } while ((*iter)->type);
 	}
      else if (input.command == "next")
 	{
	  do {
	    vector<Option*>::iterator j = val.end(); --j;
	    if (iter != j) {
	      ++iter; ++pos;
	    } else {
	      iter = val.begin(); pos = 0;
	    }
	  } while ((*iter)->type);
 	}
      else if (((*iter)->name == gettext("reload")) &&
	       (input.command == "left" || input.command == "right" || input.command == "action"))
	{
	  reload = true;
	}
      else if (input.command == "action" && (*iter)->callback != NULL)
	{
	  (*iter)->callback();
	}
      else if (input.command == "second_action")
	{
	  ExtraMenu em;

	  foreach (ExtraMenuItem& item, global->menu_items)
	    em.add_persistent_item(item);

	  em.add_item(ExtraMenuItem(gettext("Go back"), im->find_shortcut("back"),
				    boost::bind(&Options::exit, this)));

	  em.mainloop();
	}
      else if (input.command == "back" && input.mode == "general")
	{
	  exit();
	}
      else if (input.command == "left" || input.command == "right")
	{
	  if ((*iter)->name == gettext("audio device"))
	    reconfigure = true;
	  else if ((*iter)->name == gettext("sorting"))
	    reload = true;

	  if (input.command == "left") {
	    if ((*iter)->pos != 0) {
	      --(*iter)->pos;
	    } else {
	      (*iter)->pos = (*iter)->values.size() - 1;
	    }
	  }
	  else if (input.command == "right") {
	    if ((*iter)->pos != ((*iter)->values.size() - 1)) {
	      ++(*iter)->pos;
	    } else {
	      (*iter)->pos = 0;
	    }
	  }
	}
      else
	update_needed = !global->check_commands(input);
    }

  exit_loop = false;

  render->device->switch_to_layer(curlayer);
  render->device->animation_section_begin();
  render->device->animation_move(0, 0, 0, -conf->p_v_res(),  60, curlayer+1);
  render->device->animation_move(0, conf->p_v_res(), 0, 0, 60, curlayer);
  render->device->animation_fade(1,0,100,curlayer+1);
  render->device->animation_fade(0,1,100,curlayer);
  render->device->animation_section_end();

  im->restore_map();

  if (reconfigure)
    global->reconfigure_playback();

  return reload;
}

void Options::exit()
{
  exit_loop = true;
}

void Options::print_element(const string& element, const string& value,
			    bool marked, int x, int y, int max_size, int max_element_size)
{
  if (marked) {
    render->current.add(new PObj(themes->options_marked_left, x-50, y, 1, NOSCALING));
    render->current.add(new PObj(themes->options_marked, x-25, y, 90+max_size, 1));
    render->current.add(new PObj(themes->options_marked_right, x-25+90+max_size, y, 1, NOSCALING));
  }

  render->current.add(new TObj(element, "Vera/17", x, y-3, themes->options_font1,
			       themes->options_font2, themes->options_font3, 2));
  render->current.add(new TObj(value, "Vera/17", x + max_element_size + 50, y-3,
			       themes->options_font1, themes->options_font2,
			       themes->options_font3, 2));
}

void Options::print_reload_element(bool marked, int x, int y, int max_size, int max_element_size)
{

  if (marked) {
    int x_start = (conf->p_h_res()-max_size)/2;

    render->current.add(new PObj(themes->options_marked_left, x_start-50, y, 1, NOSCALING));
    render->current.add(new PObj(themes->options_marked, x_start-25, y, 90+max_size, 1));
    render->current.add(new PObj(themes->options_marked_right, x_start-25+90+max_size, y, 1, NOSCALING));
  }

  render->current.add(new TObj(gettext("reload"), "Vera/17", x, y, themes->options_font1,
			       themes->options_font2, themes->options_font3, 2));
}


void Options::print(const string& marking, int pos)
{
  render->prepare_new_image();

  render->current.add(new PObj(themes->options_background, 0, 0, 0, SCALE_FULL));

  PObj *back = new PObj(themes->startmenu_options, 25, 10, 2, NOSCALING);

  render->current.add(back);

  S_Touch::get_instance()->register_area(TouchArea(rect(back->x, back->y, back->w, back->h), back->layer,
						   boost::bind(&Options::exit, this)));

  if (themes->show_header) {
    TObj *t = new TObj(header, "Vera/28", 100, 5,
		       themes->options_header_font1, themes->options_header_font2,
		       themes->options_header_font3, 2);

    render->current.add(t);

    S_Touch::get_instance()->register_area(TouchArea(rect(t->x, t->y, t->w, t->h), t->layer, global->toggle_playlist));
  }

  int y = (conf->p_h_res()>800)?105:70;

  bool direction_down = true;
  int space = vector_lookup(val, pos)->type ? 50 : 25;
  int max_space = conf->p_v_res()-(2*y);
  int divert_pos_down = 0, divert_pos_up = 0;
  int cur_pos = 0, nr_elements = 0;
  bool direction_up_ok = true, direction_down_ok = true;

  if (header != gettext("Options"))
    y += 50;

  while (true) {
    if (!direction_down) {
      cur_pos = pos + divert_pos_up + 1;
      if (cur_pos <= val.size() - 1) {
	space += vector_lookup(val, cur_pos)->type ? 50 : 25;
	if (space > max_space) {
	  if (!direction_down_ok)
	    break;
	  space -= vector_lookup(val, cur_pos)->type ? 50 : 25;
	  direction_up_ok = false;
	  direction_down = true;
	  continue;
	}
	nr_elements += 1;
	divert_pos_up += 1;
      } else
	direction_up_ok = false;

      direction_down = true;
    } else {
      cur_pos = pos - divert_pos_down - 1;
      if (cur_pos >= 0) {
	space += vector_lookup(val, cur_pos)->type ? 50 : 25;
	if (space > max_space) {
	  if (!direction_up_ok)
	    break;
	  space -= vector_lookup(val, cur_pos)->type ? 50 : 25;
	  direction_down_ok = false;
	  direction_down = false;
	  continue;
	}

	nr_elements += 1;
	divert_pos_down += 1;
      } else
	direction_down_ok = false;

      direction_down = false;
    }

    if (!direction_up_ok && !direction_down_ok)
      break;
  }

  int max_element = 0, max_value = 0;

  bool reload_found = false;

  // calculate size
  foreach (Option *option, val) {

    if (option->name != gettext("reload") && !option->type) {

      int element_size = string_format::calculate_string_width(option->name, "Vera/17");

      foreach (string& option_value, option->values) {
	int value_size = string_format::calculate_string_width(option_value, "Vera/17");
	max_value = std::max(value_size, max_value);
      }

      max_element = std::max(element_size, max_element);
    } else if (option->name == gettext("reload")) {
      int element_size = string_format::calculate_string_width(option->name, "Vera/17");
      max_element = std::max(element_size, max_element);
      reload_found = true;
    }
  }

  if (max_element > conf->p_h_res()/2 - 20)
    max_element = conf->p_h_res()/2 - 20;

  if (max_value > conf->p_h_res()/2 - 20)
    max_value = conf->p_h_res()/2 - 20;

  int max = max_element + max_value;

  if (max == 0)
    max = 70;

  if (header != gettext("Options") && reload_found) {
    int reload_size = string_format::calculate_string_width(gettext("reload"), "Vera/17");
    int x = (conf->p_h_res()-reload_size)/2;
    print_reload_element((gettext("reload") == marking), x, y-50, max, max_element);
  }

  int x = (conf->p_h_res()-max)/2;

  int startpos = pos - divert_pos_down;

  for (int i = startpos, end = val.size(); i < end; ++i) {

    if (i - startpos > nr_elements)
      break;

    Option *o = vector_lookup(val, i);

    if (o->type) {

      y += 10;
      render->current.add(new TObj(o->name, "Vera/20", x-25, y,
				   themes->options_font1, themes->options_font2,
				   themes->options_font3, 1));
      y += 40;

    } else if (o->name != gettext("reload")) {

      string name = o->name;
      string_format::format_to_size(name, "Vera/17", max_element, false);
      string value = vector_lookup(o->values, o->pos);
      string_format::format_to_size(value, "Vera/17", max_value, false);
      if (o->callback != NULL)
	value+=" [...]";
      print_element(name, value, o->name == marking, x, y, max, max_element);
      y += 25;
    }
  }

  render->draw_and_release("options");
}

void Options::print_lcd_menu(const string& marking)
{
  if (!global->lcd_output_possible())
    return;

  std::string lcd_header = "      Options";

  // FIXME: lcd size
  //    LcdConf *lcd_conf = LcdConf::get_instance();

  vector<Option*>::iterator j = val.begin();
  while ((*j)->name != marking) {
    ++j;
  }

  std::string cur_name = "> " + (*j)->name;
  if (!(*j)->type && (*j)->values.size() > 0)
    cur_name += ": " + vector_lookup((*j)->values, (*j)->pos);
  std::string prev_name = "", next_name = "";

  vector<Option*>::iterator jend = val.end();
  --jend;

  if (val.size() == 2) {

    j = val.begin();
    if ((*j)->name != marking) {
      ++j;
    }
    if (j == jend)
      j = val.begin();
    else
      ++j;

    next_name = (*j)->name;
    if (!(*j)->type && (*j)->values.size() > 0)
      next_name += ": " + vector_lookup((*j)->values, (*j)->pos);

  } else {

    j = val.begin();
    while ((*j)->name != marking) {
      ++j;
    }

    // prev
    if (j != val.begin()) {
      --j;
      prev_name = (*j)->name;
      if (!(*j)->type && (*j)->values.size() > 0)
	prev_name += ": " + vector_lookup((*j)->values, (*j)->pos);
      ++j;
    } else {
      j = jend;
      prev_name = (*j)->name;
      if (!(*j)->type && (*j)->values.size() > 0)
	prev_name += ": " + vector_lookup((*j)->values, (*j)->pos);
      j = val.begin();
    }

    // next
    if (j != jend)
      ++j;
    else
      j = val.begin();

    next_name = (*j)->name;
    if (!(*j)->type && (*j)->values.size() > 0)
      next_name += ": " + vector_lookup((*j)->values, (*j)->pos);
  }

  global->lcd_add_output(lcd_header);
  global->lcd_add_output(prev_name.substr(0, 20));
  global->lcd_add_output(cur_name.substr(0, 20));
  global->lcd_add_output(next_name.substr(0, 20));
  global->lcd_print();
}

void Options::save()
{
  if (val.size() == 0) // don't save nothing
    return;

  std::ofstream file;

  string options_dir;
  if (!havemydir("options/", & options_dir))
    std::cerr << gettext("Could not create options directory: ") << options_dir << std::endl;

  string name = save_name;
  if (save_name.empty()) {
    name = get_english_header();
    std::cerr << gettext("Warning: module ") << name << gettext(" did not set save_name") << std::endl;
  }

  string path = options_dir + name;

  file.open(path.c_str());

  if (!file) {
    print_critical(gettext("Could not write options to file ") + path, "OPTIONS");
  } else {
    foreach (Option *option, val) {
      if (!option->type && option->name != gettext("reload"))
	file << option->english_name << ","
	     << vector_lookup(option->english_values, option->pos) << std::endl;
    }
  }

  file.close();
}

void Options::load()
{
  string name = save_name;
  if (save_name.empty()) {
    name = get_english_header();
    std::cerr << gettext("Warning: module ") << name << gettext(" did not set save_name") << std::endl;
  }

  string path = conf->p_var_data_dir() + "options/" + name;

  std::ifstream In(path.c_str());

  if(!In){
    print_critical(gettext("Could not open options file ") + path, "OPTIONS");
  } else {

    string name, value, str;
    int i;

    while (getline(In, str)) {

      string::size_type comma = str.find(",");

      name = str.substr(0,comma);
      value = str.substr(comma+1);

      foreach (Option *option, val)
	if (!option->type && option->english_name == name) {
	  i = 0;
	  foreach (string& english_value, option->english_values)
	    if (value == english_value) {
	      option->pos = i;
	      break;
	    } else
	      ++i;
	}
    }
  }
}
