#include "config.h"

#include "config.hpp"
#include "common.hpp"

#include "lyrics.hpp"

#include "common.hpp"
#include "gettext.hpp"
#include "graphics.hpp"
#include "updater.hpp"
#include "input.hpp"

#include "audio.hpp"
#include "startmenu.hpp"

#include "tinyxml.h"
#include "tinystr.h"

#include "wget_wrapper.hpp"

// saving
#include <fstream>
#include <iostream>

#include <cc++/url.h>

using std::string;

pthread_mutex_t Lyrics::singleton_mutex = PTHREAD_MUTEX_INITIALIZER;

Lyrics::Lyrics()
  : enabled(false)
{
  vecLyric.clear();

  render = S_Render::get_instance();
  BackgroundUpdater *updater = S_BackgroundUpdater::get_instance();

  updater->timer.add(TimeElement("fetchlyric",
				 boost::bind(&Lyrics::idle_status, this),
				 boost::bind(&Lyrics::check_status, this)));

  Config *conf = S_Config::get_instance();
  tmp_location = conf->p_temp_image_loc() + "/wget-temp-lyric";

  lyrics_dir = "/var/cache/mms/lyrics/";
  if (conf->p_homedir() != "/etc/mms/")
    lyrics_dir = conf->p_homedir() + "lyrics/";

  if (!file_exists(lyrics_dir)) {
    if (mkdir(lyrics_dir.c_str(), 0777) == -1)
      DebugPrint perror(dgettext("mms-audio", "Could not create directory ") + lyrics_dir,
			Print::DEBUGGING, DebugPrint::CRITICAL, "LYRICS");
  }
}

int Lyrics::idle_status()
{
  return 500; // ms
}

void Lyrics::check_status()
{
  if (enabled) {
    std::cout << "Lyrics::check_status()" << std::endl;
    Load() || search_result();
    if (buff_text.size() && !vecLyric.size())
      parse_text(buff_text);
    /*
      if (vecLyric.size())
      calculate();
    */
  }
  enabled = false;
}

string Lyrics::mainloop()
{
  return "";
}

bool Lyrics::find_lyric(string artist, string title)
{
  if (cache_artist == artist && cache_title == title)
    return true;

  cache_artist = artist;
  cache_title = title;
  vecLyric.clear();
  period = 0;
  buff_text = "";
  count = 0;
  enabled = true;
  return true;
}

bool Lyrics::Load()
{
  const int BUFFER_SIZE = 2048;
  std::ifstream file;

  file.open(filename().c_str());
  if (file.is_open())
  {
	char tmpBuff[BUFFER_SIZE];
	while(!file.eof()) {
		file.getline(tmpBuff,BUFFER_SIZE -1);
	buff_text += tmpBuff;
	buff_text += "\n";
	}
	file.close();
	std::cout << "finding on cache" << std::endl;
	return true;
  }
  return false;
}

void Lyrics::Save()
{
  std::cout << "Lyrics::Save()" << std::endl;
  if (buff_text.size())
    parse_text(buff_text);
  if (!vecLyric.size())
    return;

  std::ofstream file;
  string curLine;
  std::vector<std::string>::iterator itVecLyr;

  file.open(filename().c_str());
  if (file.is_open())
  {
    for(itVecLyr = vecLyric.begin(); itVecLyr != vecLyric.end(); itVecLyr++) {
      curLine = *(itVecLyr);
      file << curLine << std::endl;
    }
    file.close();
    std::cout << "save on cache" << std::endl;
  }
}

std::string Lyrics::filename()
{
   return (lyrics_dir + string_format::lowercase(string_format::convert(cache_artist)) + " - " + string_format::lowercase(string_format::convert(cache_title)));
}

bool Lyrics::search_result()
{
  std::cout << "finding on inet" << std::endl;
  string url = "http://api.leoslyrics.com/api_search.php?auth=MMS&artist="
	+ string_format::lowercase(string_format::convert(cache_artist))
	+ "&songtitle="
	+ string_format::lowercase(string_format::convert(cache_title));

  std::string searchxml, hid;
  int counter = 0;
  //list<pair<string, string> > genrelist;

  if (WgetWrapper::download(url, searchxml) > 0)
  {
    TiXmlDocument searchxmldoc;
    int startpos = 0;
    if((startpos = searchxml.find("<?xml")) != string::npos)
    {
      if(searchxmldoc.Parse(searchxml.substr(startpos).c_str(), 0 , TIXML_ENCODING_UTF8))
      {
	TiXmlElement* root_element = searchxmldoc.RootElement();
	TiXmlElement* element;

	  if (root_element == 0)
    	return false;

	  for (element = root_element->FirstChildElement("searchResults"); element != 0;
          element = element->NextSiblingElement("searchResults")) {
		element= element->FirstChildElement("result");
		hid = element->Attribute("hid");

	}
	std::cout << "hid :" << hid << std::endl;
/*        TiXmlElement* element;
    	element = searchxmldoc.FirstChildElement();
    	hid = element->Attribute("hid");
  	std::cout << "hid :" << hid << std::endl;*/
	return get_text(hid);
      }
    }
  }
  return false;
}

bool Lyrics::get_text(const string& hid)
{
  std::cout << "Lyrics::get_text(" << hid << ")" << std::endl;
  char hid_encoded[100];

  ost::urlEncode(hid.c_str(), hid_encoded, sizeof(hid_encoded));
  string url = "http://api.leoslyrics.com/api_lyrics.php?auth=MMS&hid=" + string(hid_encoded);


std::string searchxml;

int ret;

const char* text;

  //list<pair<string, string> > genrelist;

  if (WgetWrapper::download(url, searchxml) > 0)
  {
    TiXmlDocument searchxmldoc;
    int startpos = 0;
    if((startpos = searchxml.find("<?xml")) != string::npos)
    {
      //std::cout << "OPEN XML =" << searchxml << std::endl;
      if(searchxmldoc.Parse(searchxml.substr(startpos).c_str(), 0 , TIXML_ENCODING_UTF8))
      {
	TiXmlNode* root_element = searchxmldoc.RootElement();
	TiXmlNode* element;
	TiXmlNode* response;

	if (root_element == 0)
    	  return false;

	response = root_element->FirstChildElement("response");
	ret = conv::atoi(response->ToElement()->Attribute("code"));

	if (ret == 0) {
	  for (element = root_element->FirstChildElement("lyric"); element != 0;
          element = element->NextSiblingElement("lyric")) {
	    element = element->FirstChildElement("text");
	    element = element->FirstChild();
	    buff_text = string_format::trim(element->Value());
            Save();
            std::cout << "ending Lyrics::get_text(" << hid << ")" << std::endl;
            return true;
          }
        }
      }
    }
  }
  return false;
}

bool Lyrics::parse_text (const string& text, const string& nsec)
{
  int startpos = -1;
  int currentpos = 0;
  string line, temp;
  temp = text;

  vector<std::string> vecTmp;
  vecTmp.clear();

  /*for (int i=0; i < 14; i++)
	 vecTmp.push_back(" ");
*/

  std::string::size_type pos = temp.find(nsec);
  if (pos == std::string::npos && nsec != "\n")
    return parse_text(text, "\n");

  while (pos != std::string::npos) {
    vecTmp.push_back(string_format::trim(text.substr(startpos + 1 , pos - startpos - 1 )));
    temp.replace(pos, 1, "|");
    startpos = pos;
    pos = temp.find(nsec);
  }

  vecTmp.push_back(string_format::trim(text.substr(startpos + 1 , pos - startpos - 1 )));

  /*for (int i=0; i < 7; i++)
	 vecTmp.push_back(" "); */

  vecLyric = vecTmp;
  std::cout << "Lyrics::parse_text(text) loaded "<< vecLyric.size() << " lines" << std::endl;
  return true;
}


bool Lyrics::is_enabled()
{
  return (vecLyric.size() ? true : false);
}

