#include "config.h"

#include "libfspp.hpp"

#include "module.hpp"

#include "input.hpp"
#include "config.hpp"
#include "common.hpp"
#include "renderer.hpp"
#include "text_object.hpp"

#include "updater.hpp"

#include "boost.hpp"

#include <iostream>

using std::vector;
using std::list;
using std::string;

Module::Module()
  : search_mode(false), search_str(""), offset(0),
    search_help_offset(-1), search_need_cleanup(false),
    search_top_size(35), search_depth(gettext("current folder")),
    overlay_search("search marker"), id(1), visible(false),
    loaded_correctly(false), reload_needed(false)
{
  input_master = S_InputMaster::get_instance();
  themes = S_Themes::get_instance();
  render = S_Render::get_instance();
  conf = S_Config::get_instance();
  global = S_Global::get_instance();
  if (conf->p_media())
    cd = S_Cd::get_instance();

  int size = 55;

  string all_chars = global->get_valid_keys();

  letters.clear();

  string spaced_keys = "";
  vector<int> group_start;

  string font = "Vera/22";
  int start_pos = 51;

  if (conf->p_h_res() == 720) {
    font = "Vera/16";
    start_pos = 16;
  }

  if (conf->p_h_res() == 800) {
    font = "Vera/18";
    start_pos = 16;
  }

  for (int i = 0; i < all_chars.size(); ++i) {
    if (all_chars[i] != '|') {
      spaced_keys += (all_chars.substr(i,1) + " ");
    }

    if (all_chars[i] == '|' || i == all_chars.size()-1) {

      spaced_keys = spaced_keys.substr(0, spaced_keys.size()-1);

      int letters_width = string_format::calculate_string_width(spaced_keys, font);

      spaced_keys = "";
      start_pos += 5 + letters_width;

      group_start.push_back(start_pos);
    }
  }

  int group_pos = 0;

  for (int i = 0; i < all_chars.size(); ++i) {
    string cur_char = all_chars.substr(i, 1);
    if (cur_char != "|") {
      letters.push_back(size);
      int letter_space = string_format::calculate_string_width(cur_char, font);
      size += (letter_space-1);
    } else {
      size = vector_lookup(group_start, group_pos);
      ++group_pos;
    }
  }
}

bool Module::init()
{
  return true;
}

void Module::folder_difference(const vector<string>& new_list,
			       const vector<string>& old_list,
			       vector<string>& new_files,
			       vector<string>& removed_files)
{
  // find what is new
  foreach (const string& new_item, new_list) {
    bool found_match = false;
    foreach (const string& old_item, old_list) {
      if (new_item == old_item) {
	found_match = true;
	break;
      }
    }
    if (!found_match)
      new_files.push_back(new_item);
  }

  // ... and what is removed
  foreach (const string& old_item, old_list) {
    bool found_match = false;
    foreach (const string& new_item, new_list) {
      if (old_item == new_item) {
	found_match = true;
	break;
      }
    }
    if (!found_match)
      removed_files.push_back(old_item);
  }
}

int Module::check_metadata()
{
  return 3000; // ms
}

int Module::check_search_letter()
{
  return 200; // ms
}

void Module::print_marked_search_letter()
{
  if (conf->p_input() != "lirc" && conf->p_input() != "evdev")
    return;

  if (search_mode) {

    search_need_cleanup = true;

    render->wait_and_aquire();

    int y_offset = 50;

    if (overlay_search.elements.size() > 0)
      overlay_search.partial_cleanup();

    if (time(0)-conf->p_last_key() <= 2 && search_help_offset != -1)
      overlay_search.add(new RObj(vector_lookup(letters, search_help_offset), 60+y_offset,
				  25, 25, 160, 150, 90, 255, 4));

    string keys = global->get_valid_keys();
    string spaced_keys = "";

    int start_pos = 55;

    if (conf->p_h_res() <= 800)
      start_pos = 20;

    for (int i = 0; i < keys.size(); ++i) {
      if (keys[i] != '|')
	spaced_keys += (keys.substr(i,1) + " ");

      if (keys[i] == '|' || i == keys.size()-1) {

	spaced_keys = spaced_keys.substr(0, spaced_keys.size()-1);

	string font = "Vera/22";

	if (conf->p_h_res() == 720)
	  font = "Vera/16";
	else if (conf->p_h_res() == 800)
	  font = "Vera/18";

	int letters_width = string_format::calculate_string_width(spaced_keys, font);

	overlay_search.add(new RObj(start_pos-4, 60+y_offset, letters_width, 25,
				    20, 30, 20, 255, 3));

	overlay_search.add(new TObj(spaced_keys, font,
				    start_pos, 60+y_offset-5, themes->audio_font1,
				    themes->audio_font2, themes->audio_font3, 5));

	spaced_keys = "";
	start_pos = start_pos + 5 + letters_width;
      }
    }

    render->draw_and_release("search marker", true);
  }
}

void Module::clean_up_search_print()
{
  if (search_need_cleanup) {
    render->wait_and_aquire();
    overlay_search.cleanup();
    render->image_mut.leaveMutex();
    search_need_cleanup = false;
  }
}

void Module::print_range_top(int int_position, int list_size, int element_size,
			     int& y, int& start, int& range)
{
  y = 105;

  if (conf->p_v_res() == 405)
    y = 85;

  if (search_mode)
    y += search_top_size;

  int buttom = 90;

  if (conf->p_v_res() == 405)
    buttom = 65;

  range = (conf->p_v_res()-y-buttom)/element_size;

  int width = conf->p_h_res()-2*55;
  int height = range*element_size+25;

  if (search_mode)
    render->current.add(new RObj(55, y-15, width, height,
				 themes->general_rectangle_color1, themes->general_rectangle_color2,
				 themes->general_rectangle_color3, themes->general_rectangle_color4, 1));
  else
    render->current.add(new RObj(55, y-15, width, height,
				 themes->general_rectangle_color1, themes->general_rectangle_color2,
				 themes->general_rectangle_color3, themes->general_rectangle_color4, 1));

  start = int_position - (range/2);

  // special cases
  if (range < list_size) {
    if (int_position >= list_size - (range/2)) // end
      start = list_size - range;
    else if ((range/2) >= int_position) // beginning
      start = 0;
  } else { // playlist smaller than range
    start = 0;
    range = list_size;
  }
}

// ====================================================================
//
// dir
//
// ====================================================================

// Add a file to the list
Simplefile Module::addsimplefile(const string& name, const MyPair& filetype, bool from_media)
{
  Simplefile r;
  r.id = ++id;
  r.path = name;

  string::size_type pos = name.rfind("/");
  if (pos != string::npos)
    r.name = name.substr(pos+1, name.size()-((pos+1)+(filetype.first.size()+1)));
  else
    r.name = name.substr(0, name.size()-(filetype.first.size()+1));

  if (conf->p_convert())
    r.name = string_format::convert(r.name);

  r.lowercase_name = string_format::lowercase(r.name);

  if (from_media)
    r.type = "media-file";
  else
    r.type = filetype.second;

  return r;
}
