#include "guiwindowmanager.hpp"

using std::string;

pthread_mutex_t pymms::gui::GUIWindowManager::singleton_mutex = PTHREAD_MUTEX_INITIALIZER;

pymms::gui::GUIWindowManager::GUIWindowManager() : m_windowOverlay(Overlay("WindowManager"))
{}

bool pymms::gui::GUIWindowManager::onAction(const string& strAction)
{
  if(!m_vecWindows.empty())
    return  m_vecWindows.front()->onAction(strAction);
  else
     return false;
}

void pymms::gui::GUIWindowManager::registerWindow(GUIWindow* pWindow)
{
  for (std::vector<GUIWindow*>::iterator i = m_vecWindows.begin(),
       end = m_vecWindows.end(); i != end; i++) {
    if((*i) == pWindow)
    {
      return;
    }
  }

  m_vecWindows.push_back(pWindow);
}

void pymms::gui::GUIWindowManager::unregisterWindow(GUIWindow* pWindow)
{
  Render* pRenderer = S_Render::get_instance();

  for (std::vector<GUIWindow*>::iterator i = m_vecWindows.begin(),
       end = m_vecWindows.end(); i != end; i++) {
    if((*i) == pWindow)
    {
      if(pWindow->getOverlay())
      {
          pRenderer->wait_and_aquire();
          m_windowOverlay.cleanup();
  	  pRenderer->image_mut.leaveMutex();
      }

      m_vecWindows.erase(i);

      render();
      return;
    }
  }
}

void pymms::gui::GUIWindowManager::setActiveWindow(GUIWindow* pWindow)
{
  for (std::vector<GUIWindow*>::iterator i = m_vecWindows.begin(),
       end = m_vecWindows.end(); i != end; i++) {
    if((*i) == pWindow)
    {
      m_vecWindows.erase(i);
      m_vecWindows.insert(m_vecWindows.begin(), pWindow);

      render();
      return;
    }
  }
}

void pymms::gui::GUIWindowManager::setActiveWindow(int iWindowId)
{
  for (std::vector<GUIWindow*>::iterator i = m_vecWindows.begin(),
       end = m_vecWindows.end(); i != end; i++) {
    if((*i)->getId() == iWindowId)
    {
      GUIWindow* pWindow = (*i);

      m_vecWindows.erase(i);
      m_vecWindows.insert(m_vecWindows.begin(), pWindow);

      render();
      return;
    }
  }
}

pymms::gui::GUIWindow* pymms::gui::GUIWindowManager::getActiveWindow()
{
  if(!m_vecWindows.empty())
    return m_vecWindows.front();
  else
    return 0;
}

pymms::gui::GUIWindow* pymms::gui::GUIWindowManager::getWindow(int iWindowId)
{
  for (std::vector<GUIWindow*>::iterator i = m_vecWindows.begin(),
       end = m_vecWindows.end(); i != end; i++) {
    if((*i)->getId() == iWindowId)
    {
      return (*i);
    }
  }

  return 0;
}

void pymms::gui::GUIWindowManager::render()
{
  Render* pRenderer = S_Render::get_instance();
  GUIWindow* pWindow = (!m_vecWindows.empty()?m_vecWindows.front():0);

  if(pWindow && !pWindow->getOverlay())
  {
    pRenderer->prepare_new_image();

    pWindow->render();

    pRenderer->draw_and_release("Window");
  }
  else if(pWindow && pWindow->getOverlay())
  {
     pRenderer->wait_and_aquire();

     if (m_windowOverlay.elements.size() > 0)
        m_windowOverlay.partial_cleanup();

     pWindow->render(&m_windowOverlay);

     pRenderer->draw_and_release("Window", true);
  }
}

void pymms::gui::GUIWindowManager::sendMessage(GUIMessage& message)
{ 
  GUIWindow* pWindow = (!m_vecWindows.empty()?m_vecWindows.front():0);

  if(pWindow)
     pWindow->onMessage(message);
}
