#include "search.hpp"

#include "global.hpp"
#include "graphics.hpp"
#include "extra_menu.hpp"
#include "busy_indicator.hpp"
#include "updater.hpp"
#include "touch.hpp"

#include "boost.hpp"

using std::string;

pthread_mutex_t Search::singleton_mutex = PTHREAD_MUTEX_INITIALIZER;

SearchModule::SearchModule(const string& n, const list_gen_func& sgl, const enter_func& e_f)
  : list_gen(sgl), enter(e_f), name(n)
{}

Search::Search()
  : search_word(""), search_module(gettext("all")), pos(-1), em("", true), exit_loop(false)
{
  em.set_drop_down_pos(conf->p_h_res()-250+85, 85);
  em.add_item(ExtraMenuItem(gettext("all"), "", boost::bind(&Search::set_module, this, gettext("all"))));
  em.add_exit_input(Input("left", "general"));
}

void Search::set_module(const string& mod)
{
  search_module = mod;
}

void Search::find_element_and_do_action(const result& element)
{
  int p = 0;
  foreach (result& f, results) {
    if (f == element) {
      pos = p;
      action();
      break;
    } else
      ++p;
  }
}

void Search::print_element(const result& r, const result& position, int& y)
{
  if (r.module != cur_module) {

    if (y != 135)
      y += 10;

    render->current.add(new TObj(r.module, "Vera/18", 85, y,
				 themes->search_module_font1, themes->search_module_font2,
				 themes->search_module_font3, 3));
    y += 35;
    cur_module = r.module;
  }

  string name = r.name;

  string_format::format_to_size(name, "Vera/16", conf->p_h_res()-2*70-50, true);

  PObj *p = new PObj(themes->search_marked_left, 70, y, 2, NOSCALING);

  S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, conf->p_h_res()-(2*70)-50, p->h), p->layer, boost::bind(&Search::find_element_and_do_action, this, r)));

  if (r == position) {
    render->current.add(p);
    render->current.add(new PObj(themes->search_marked, 70+25, y, conf->p_h_res()-(2*70)-50, 2));
    render->current.add(new PObj(themes->search_marked_right, 70+25+conf->p_h_res()-(2*70)-50, y, 2,
				 NOSCALING));

    render->current.add(new TObj(name, "Vera/16", 100, y,
				 themes->search_marked_font1, themes->search_marked_font2,
				 themes->search_marked_font3, 3));
  } else {
    render->current.add(new TObj(name, "Vera/16", 100, y,
				 themes->search_font1, themes->search_font2,
				 themes->search_font3, 3));
  }
}

void Search::print(const string& search_w)
{
  render->prepare_new_image();

  render->current.add(new PObj(themes->search_background, 0, 0, 0, SCALE_FULL));

  PObj *back = new PObj(themes->general_search, 25, 10, 2, NOSCALING);

  render->current.add(back);

  S_Touch::get_instance()->register_area(TouchArea(rect(back->x, back->y, back->w, back->h), back->layer,
						   boost::bind(&Search::exit, this)));

  if (themes->show_header) {
    TObj *t = new TObj(gettext("Search"), "Vera/28", 100, 5,
		       themes->search_header_font1, themes->search_header_font2,
		       themes->search_header_font3, 2);

    render->current.add(t);

    S_Touch::get_instance()->register_area(TouchArea(rect(t->x, t->y, t->w, t->h), t->layer, global->toggle_playlist));
  }

  int x = 85, y = 85, max_x = conf->p_h_res()-250;

  if (pos == -1)
    render->current.add(new RObj(x-5, y-3, max_x+10, 25+6, 0, 0, 0, 215, 2));

  string sw = search_w;

  string s = gettext("Search: ");
  int x_size = string_format::calculate_string_width(s, "Vera/17");
  string_format::format_to_size(sw, "Vera/17", max_x-x_size, true);

  render->current.add(new TObj(s + sw, "Vera/17", x, y, themes->search_font1,
			       themes->search_font2, themes->search_font3, 3));

  render->current.add(new TObj(search_module, "Vera/17", x+max_x+25, y, themes->search_font1,
			       themes->search_font2, themes->search_font3, 3));

  y += 50;

  if (conf->p_input() != "keyboard")
    y += 35;

  if (results.size() > 0) {

    int buttom = 90;

    if (conf->p_v_res() == 405)
      buttom = 65;

    bool direction_down = true;
    int space = 63;
    int max_space = conf->p_v_res()-(y+buttom);
    int divert_pos_down = 0, divert_pos_up = 0;
    int cur_pos = 0, nr_elements = 0;
    bool direction_up_ok = true, direction_down_ok = true;
    bool down_failed = false, up_failed = false;

    // the value (bool) is dummy, we just need a fast lookup mechanism
    std::map<string, bool> available_modules;
    if (pos != -1)
      available_modules.insert(std::make_pair(vector_lookup(results, pos).module, true));
    else
      available_modules.insert(std::make_pair(vector_lookup(results, 0).module, true));

    string module;

    int last_add;

    while (true) {

      if (!direction_down) {
	cur_pos = pos + divert_pos_up + 1;
	if (cur_pos <= results.size() - 1) {
	  module = vector_lookup(results, cur_pos).module;
	  if (available_modules.find(module) == available_modules.end()) {
	    space += 63;
	    available_modules.insert(std::make_pair(module, true));
	  } else {
	    space += 28;
	  }

	  if (space > max_space) {
	    if (down_failed)
	      break;
	    else {
	      up_failed = true;
	      direction_down = true;
	      continue;
	    }
	  } else {
	    nr_elements += 1;
	    divert_pos_up += 1;
	  }
	} else {
	  up_failed = true;
	  direction_up_ok = false;
	}

	direction_down = true;
      } else {
	cur_pos = pos - divert_pos_down - 1;
	if (cur_pos >= 0) {
	  module = vector_lookup(results, cur_pos).module;
	  if (available_modules.find(module) == available_modules.end()) {
	    space += 63;
	    last_add = 63;
	    available_modules.insert(std::make_pair(module, true));
	  } else {
	    space += 28;
	    last_add = 28;
	  }

	  if (space > max_space) {
	    if (up_failed)
	      break;
	    else {
	      space -= last_add;
	      down_failed = true;
	      direction_down = false;
	      continue;
	    }
	  } else {
	    nr_elements += 1;
	    divert_pos_down += 1;
	  }
	} else {
	  down_failed = true;
	  direction_down_ok = false;
	}

	direction_down = false;
      }

      if (!direction_up_ok && !direction_down_ok)
	break;
    }

    cur_module = "";

    int startpos = pos - divert_pos_down;

    // pos == -1
    if (startpos < 0)
      startpos = 0;

    render->current.add(new RObj(55, y-10, conf->p_h_res()-2*55, max_space+20,
				 themes->general_rectangle_color1, themes->general_rectangle_color2,
				 themes->general_rectangle_color3, themes->general_rectangle_color4, 1));

    result cur_element;

    if (pos != -1)
      cur_element = vector_lookup(results, pos);

    for (int i = startpos, end = results.size(); i < end; ++i) {

      if (i - startpos > nr_elements)
	break;

      print_element(vector_lookup(results, i), cur_element, y);

      y += 28;
    }
  }

  render->draw_and_release("search");
}

void Search::register_module(const SearchModule& s)
{
  if (s.name != gettext("all")) {
    modules.push_back(s);
    em.add_item(ExtraMenuItem(s.name, "", boost::bind(&Search::set_module, this, s.name)));
  } else
    // FIXME: complain loudly
    ;

  print_debug("registering module:" + s.name, "SEARCH");
}

class SearchDialogFunctor
{
 private:
  void (*fpt)(const string&, int, Overlay&);
  Overlay o;

 public:

  SearchDialogFunctor(void(*_fpt)(const string&, int, Overlay&))
    : o(Overlay("fixed size dialog"))
  { fpt=_fpt; };

  ~SearchDialogFunctor()
  {
    Render *render = S_Render::get_instance();
    render->image_mut.enterMutex();
    o.cleanup();
    render->image_mut.leaveMutex();
  }

  void call(const string& r)
  { return (*fpt)(gettext("Search: ") + r, 450, o); };
};

void Search::action()
{
  result cur = vector_lookup(results, pos);
  foreach (SearchModule& module, modules)
    if (cur.module == module.name) {
      module.enter(cur.id);
      break;
    }
}

void Search::exit()
{
  exit_loop = true;
}

string Search::mainloop()
{
  Input input;

  print(search_word);

  std::list<Input> exit_inputs;
  exit_inputs.push_back(Input("back", "general"));
  exit_inputs.push_back(Input("right", "general"));
  exit_inputs.push_back(Input("next", "general"));

  bool in_select_module = false;

  S_BusyIndicator::get_instance()->idle();

  search_mode = true;

  std::list<string> empty;
  empty.push_back("");

  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();

  bool update_needed = true;

  while (!exit_loop)
    {
      if (pos == -1 && !in_select_module) {

	string old_search_word = search_word;

	input_master->set_map("search");

	print_marked_search_letter();

	screen_updater->trigger.add(TriggerElement("search marker", empty,
						   boost::bind(&Search::print_marked_search_letter, this),
						   boost::bind(&Search::clean_up_search_print, this)));
	screen_updater->timer.add(TimeElement("search marker",
					      boost::bind(&Search::check_search_letter, this),
					      boost::bind(&Search::print_marked_search_letter, this)));

	std::pair<bool, Input> res = input_master->device->generate_string(search_word, search_help_offset,
									   boost::bind(&Search::print, this, _1),
									   exit_inputs);

	screen_updater->trigger.del("search marker");
	screen_updater->timer.del("search marker");

	if (res.first) {
	  results.clear();
	  foreach (SearchModule& module, modules)
	    if (search_module == module.name || search_module == gettext("all")) {
	      std::vector<std::pair<string, int> > temp = module.list_gen(search_word);
	      for (std::vector<std::pair<string, int> >::iterator ti = temp.begin(), tiend = temp.end();
		   ti != tiend; ++ti) {
		result r;
		r.module = module.name;
		r.name = ti->first;
		r.id = ti->second;
		results.push_back(r);
	      }
	    }

	  if (results.size() > 0) {
	    input_master->set_map("default");
	    pos = 0;
	  } else {
	    DialogWaitPrint pdialog(gettext("No search results found"), 2000);
	    continue;
	  }

	} else {

	  if (res.second.command == "exit" || res.second.command == "back") {
	    search_word = old_search_word;
	    exit_loop = true;
	  } else if (res.second.command == "right") {
	    in_select_module = true;
	    continue;
	  } else if (res.second.command == "next" && results.size() > 0) {
	    input_master->set_map("default");
	    pos = 0;
	    continue;
	  } else
	    input = res.second;
	}

	continue;

      } else if (pos == -1 && in_select_module) {

	em.mainloop();

	S_BusyIndicator::get_instance()->idle();

	print(search_word);
	in_select_module = false;
	continue;

      } else {
	if (update_needed)
	  print(search_word);

	update_needed = true;
      }

      input = input_master->device->get_input();

      if (input.key == "touch_input") {
	S_Touch::get_instance()->run_callback();
	continue;
      }

      if (input.command == "prev")
	{
	  if (pos != -1)
	    --pos;

	  // if we go up again we need to draw again since otherwise the first
	  // element will be marked
	  if (pos == -1)
	    print(search_word);
 	}
      else if (input.command == "next")
 	{
	  if (pos != results.size()-1)
	    ++pos;
 	}
      else if (input.command == "action")
	{
	  action();
	}
      else if (input.command == "back" && input.mode == "general")
	{
	  exit();
	}
      else
	update_needed = !global->check_commands(input, true);
    }

  exit_loop = false;
  search_mode = false;

  return "";
}
