// used for #ifdef
#include "config.h"

#include "startmenu.hpp"

//Input
#include "input.hpp"

#include "gettext.hpp"
#include "renderer.hpp"
#include "config.hpp"
#include "graphics.hpp"
#include "updater.hpp"
#include "plugins.hpp"
#include "input.hpp"
#include "busy_indicator.hpp"
#include "shutdown.hpp"
#include "extra_menu.hpp"

#include "theme.hpp"
#include "search.hpp"

#include "touch.hpp"

#include "boost.hpp"

#ifdef use_inotify
#include "notify.hpp"
#endif

// ostringstream
#include <sstream>

#include <iostream>

using std::list;
using std::string;
using std::vector;

pthread_mutex_t Startmenu::singleton_mutex = PTHREAD_MUTEX_INITIALIZER;

startmenu_item::startmenu_item(const string& n, const string& a, const string& p, int i,
			       const callback_func& f)
  : name(n), action(a), pic(p), id(i), callback(f)
{
}

void Startmenu::StartmenuOpts::set_options()
{
  // header
  header = gettext("Options");

  // values
  vector<string> empty_values;

  // global options
  if (conf->p_cdpath().size() > 1 || go->audio_device->values.size() > 1) {
    g_o = new Option(true, gettext("Global Options"), "Global Options", 0,
		     empty_values, empty_values);
    val.push_back(g_o);

    if (conf->p_cdpath().size() > 1)
      val.push_back(go->cd_device);

    if (go->audio_device->values.size() > 1)
      val.push_back(go->audio_device);
  }

  Plugins *plugins = S_Plugins::get_instance();
  foreach (FeaturePlugin *plugin, plugins->fp_data)
    if (plugin->module && plugin->module->get_opts()) {

      plugin->module->get_opts()->load();

      std::vector<Option*> plugin_values = plugin->module->get_opts()->values();

      // FIXME: little weak check since not everyone has both sorting and reload
      if (plugin_values.size() < 3)
	continue;

      val.push_back(new Option (true, plugin->module->get_opts()->get_header(),
				plugin->module->get_opts()->get_english_header(),
				0, empty_values, empty_values));

      for (unsigned int i = 0; i < plugin_values.size(); ++i) {
	string name = plugin_values[i]->name;
        if (name != gettext("reload") && name != gettext("sorting") &&
	    name != gettext("cd device") && name != gettext("audio device"))
	  val.push_back(plugin_values[i]);
      }
    }

  load();
}

Startmenu::StartmenuOpts::~StartmenuOpts()
{
  // global
  delete g_o;
}

void Startmenu::StartmenuOpts::save()
{
  // intentionally left blank
}

void Startmenu::StartmenuOpts::load()
{
  // intentionally left blank
}

void Startmenu::print_lcd_menu(const startmenu_row& row, const startmenu_item& current_item)
{
  if (!global->lcd_output_possible())
    return;

  global->lcd_add_output("     Startmenu");
  global->lcd_add_output("");
  global->lcd_add_output(string_format::pretty_print_string_copy(row.name + ":", 20));
  global->lcd_add_output(string_format::pretty_print_string_copy(current_item.name, 20));
  global->lcd_print();
}

string Startmenu::mainloop()
{
  InputMaster *input_master = S_InputMaster::get_instance();

  Input input;

  input_master->set_map("default");

  list<startmenu_row>::iterator startmenu_iter = startmenu.begin();
  list<startmenu_item>::iterator startmenu_iter_row = startmenu_iter->items.begin();

  BusyIndicator *busy_indicator = S_BusyIndicator::get_instance();
  Shutdown *sd = S_Shutdown::get_instance();

  Plugins *plugins = S_Plugins::get_instance();
  foreach (FeaturePlugin *plugin, plugins->fp_data)
    plugin->module->startup_updater();

  opts.set_options();

  // updater
  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();

  screen_updater->start();
  screen_updater->activate();

  BackgroundUpdater *background_updater = S_BackgroundUpdater::get_instance();

  background_updater->start();
  background_updater->activate();

#ifdef use_inotify
  S_Notify::get_instance()->start();
#endif

  bool update_needed = true;
  while (!exit_loop)
    {
      if (update_needed) {
	print(*startmenu_iter, *startmenu_iter_row);
	print_lcd_menu(*startmenu_iter, *startmenu_iter_row);
      }

      input = input_master->device->get_input_busy_wrapped(busy_indicator);

      if (sd->is_enabled()) {
      	sd->cancel();
      	continue;
      }

      update_needed = true;

      if (input.key == "touch_input") {
	S_Touch::get_instance()->run_callback();
	continue;
      }

      if (input.command == "prev")
 	{
 	  if (startmenu_iter != startmenu.begin()) {
	    --startmenu_iter;
	    startmenu_iter_row = startmenu_iter->items.begin();
	  }
 	}
      else if (input.command == "next")
 	{
 	  list<startmenu_row>::iterator j = startmenu.end(); --j;
 	  if (startmenu_iter != j) {
 	    ++startmenu_iter;
	    startmenu_iter_row = startmenu_iter->items.begin();
	  }
 	}
      else if (input.command == "left")
 	{
 	  if (startmenu_iter_row != startmenu_iter->items.begin())
 	    --startmenu_iter_row;
 	}
      else if (input.command == "right")
 	{
 	  list<startmenu_item>::iterator j = startmenu_iter->items.end(); --j;
 	  if (startmenu_iter_row != j)
 	    ++startmenu_iter_row;
 	}
      else if (input.command == "second_action")
	{
	  ExtraMenu em;

	  foreach (ExtraMenuItem& item, global->menu_items)
	    em.add_persistent_item(item);

	  em.mainloop();
	}
      else if (input.command == "action" && startmenu_iter_row->action != "quit")
 	{
	  print_debug("calling feature plugin mainloop");
	  foreach (startmenu_row& row, startmenu)
	    foreach (startmenu_item& item, row.items) {
	      if (item.action == startmenu_iter_row->action) {
		print_debug("Enter plugin action " + row.name);
		item.callback();

		// set the input map back to startmenu
		input_master->set_map("default");
		break;
	    }
	  }
	}
      else if (input.command == "action" && startmenu_iter_row->action == "quit")
	{
	  quit_func();
	}
      else
	// FIXME: ?
	// make back work for playlist commands
	//input_master->cur_map = &input_master->standard;
	update_needed = !global->check_commands(input);
    }
  return "quit";
}

void Startmenu::print(const startmenu_row& row, const startmenu_item& current_item)
{
  Render *render = S_Render::get_instance();
  Themes *themes = S_Themes::get_instance();

  render->prepare_new_image();

  render->current.add(new PObj(themes->startmenu_background, 0, 0, 0, SCALE_FULL));

  int pos = 0;
  foreach (startmenu_row& cur_row, startmenu) {
    if (cur_row.name == row.name)
      break;
    ++pos;
  }

  int y_top = 120;
  int y_bottom = 40;
  int x = 0;

  if (conf->p_h_res() == 720)
    y_top = 40;

  int max_y = 51;

  int nr_items = startmenu.size();
  if (max_y*startmenu.size()+y_top+y_bottom > conf->p_v_res())
    nr_items = (conf->p_v_res()-y_top-y_bottom)/max_y;

  int box_height = max_y*nr_items;

  int y = (conf->p_v_res()-(box_height))/2;

  int start = pos - (nr_items/2);

  // special cases
  if (nr_items < startmenu.size()) {
    if (pos >= startmenu.size() - (nr_items/2)) // end
      start = startmenu.size() - nr_items;
    else if ((nr_items/2) >= pos) // beginning
      start = 0;
  } else { // list smaller than nr_items
    start = 0;
  }

  int i = 0;
  list<startmenu_row>::const_iterator iter = startmenu.begin();
  for (int t = 0; t < start; ++t)
    ++iter;
  for (; iter != startmenu.end() && i < nr_items; ++iter, ++i)
    {
      if (iter->name == row.name) {

	foreach (const startmenu_item& item, iter->items) {

	  if (current_item.name == item.name) {

	    render->current.add(new PObj("startmenu/marked-left.png", 65, y, 1, NOSCALING));

	    if (x == 70)
	      render->current.add(new PObj("startmenu/marked2.png", 90, y, conf->p_h_res()-220+x-50, 1));
	    else if (x == 140)
	      render->current.add(new PObj("startmenu/marked3.png", 90, y, conf->p_h_res()-220+x-50, 1));
	    else
	      render->current.add(new PObj("startmenu/marked.png", 90, y, conf->p_h_res()-220+x-50, 1));

	    render->current.add(new PObj("startmenu/marked-right.png", 90+conf->p_h_res()-220+x-50, y, 1, NOSCALING));

	    render->current.add(new TObj(item.name, "Vera/22", 75, y+5,
					 themes->startmenu_font1, themes->startmenu_font2,
					 themes->startmenu_font3, 2));
	  }
	  PObj *p = new PObj(item.pic, conf->p_h_res()-220+x, y, 2, NOSCALING);
	  render->current.add(p);
	  S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, p->w, p->h), p->layer, item.callback));
	  x += 70;
	}
      } else {
	render->current.add(new TObj(iter->name, "Vera/22", 75, y+5,
				     themes->startmenu_font1, themes->startmenu_font2,
				     themes->startmenu_font3, 2));
	foreach (const startmenu_item& item, iter->items) {
	  PObj* p= new PObj(item.pic, conf->p_h_res()-220+x, y, 2, NOSCALING);
	  render->current.add(p);
	  S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, p->w, p->h), p->layer, item.callback));
	  x += 70;
	}
      }
      x = 0;
      y += 51;
    }

  // render the logo in the startmenu
  render->current.add(new PObj("logo.png", conf->p_h_res()-85, conf->p_v_res()-41, 2, NOSCALING));

  render->draw_and_release("startmenu");
}

Startmenu::Startmenu()
  : exit_loop(false)
{
  generate_startmenu();
  conf = S_Config::get_instance();
  global = S_Global::get_instance();
}

void search_func()
{
  S_Search::get_instance()->mainloop();
}

void options_func()
{
  S_BusyIndicator::get_instance()->idle();
  S_Startmenu::get_instance()->opts.mainloop();
}

void open_close_func()
{
  S_Cd::get_instance()->openclose_cddrive();
}

void themes_func()
{
  S_BusyIndicator::get_instance()->idle();
  S_Theme::get_instance()->mainloop();
}

void Startmenu::quit_func()
{
  ExtraMenu em(gettext("Do you really want to exit?"));

  if (conf->p_show_quit())
    em.add_item(ExtraMenuItem(gettext("Quit"), "", NULL));
  if (!conf->p_shutdown_script().empty())
    em.add_item(ExtraMenuItem(gettext("Halt System"), "", NULL));
  em.add_item(ExtraMenuItem(gettext("Cancel"), "", NULL));

  int ret = em.mainloop();

  if (ret == 0 && conf->p_show_quit()) {
    conf->s_shutdown_script("");
    exit_loop = true;
  } else if (!conf->p_shutdown_script().empty() && (ret == 1 && conf->p_show_quit() || ret == 0 && !conf->p_show_quit()))
    exit_loop = true;
}

void Startmenu::generate_startmenu()
{
  Themes *themes = S_Themes::get_instance();
  Config *conf = S_Config::get_instance();

  // generate menues.
  list<startmenu_row> start_menu;
  startmenu_row row;

/*
  Category stored in plugin_name() is what menu displays when unselected
  ie Audio, Video, Midi, Picture, EPG, TV, NET...
  priority was for column dynamic allocation (0..2) (unused)
*/

  Plugins *plugins = S_Plugins::get_instance();
  foreach (FeaturePlugin *plugin, plugins->fp_data) {
    if (plugin->module != 0 && plugin->module->init()) {
      print_debug("Registering plugin features for " + plugin->plugin_name());
      row.name = plugin->plugin_name();
      row.items = plugin->items();
      start_menu.push_back(row);
    }
  }

  row.items.clear();

  // search button
  row.items.push_back(startmenu_item(gettext("Search"), "search",
				     themes->startmenu_search, 0, &search_func));

  // options button
  row.items.push_back(startmenu_item(gettext("Options"), "options",
				     themes->startmenu_options, 0, &options_func));

  // themes button
  row.items.push_back(startmenu_item(gettext("Themes"), "themes",
				     themes->startmenu_themes, 1, &themes_func));

  if (row.items.size() != 0) {
    row.name = gettext("Search");
    start_menu.push_back(row);
    row.items.clear();
  }

  // Open/Close Tray button
  if (conf->p_openclose()) {
    row.items.push_back(startmenu_item(gettext("Open/Close Tray"), "openclose",
				       themes->startmenu_open_close, 0, &open_close_func));

    row.name = gettext("Open/Close Tray");
    start_menu.push_back(row);
    row.items.clear();
  }

  // quit button
  row.name = gettext("Quit");
  row.items.push_back(startmenu_item(gettext("Quit"), "quit", themes->startmenu_quit, 0,
				     boost::bind(&Startmenu::quit_func, this)));

  start_menu.push_back(row);
  row.items.clear();

  startmenu = start_menu;
}
